/*----------------------------------------------------------------------
  File    : Font3D.java
  Contents: simple 3D drawing of strings
  Author  : Christian Borgelt
  History : 2004.06.03 file created
            2004.06.05 adapted to translation and scaling
            2004.06.06 clipping functions added
            2007.02.07 javadoc added
----------------------------------------------------------------------*/
package draw;

import java.awt.Graphics;
import java.io.Serializable;

/*--------------------------------------------------------------------*/
/** Class for a simple 3D drawing of strings.
 *  @author Christian Borgelt
 *  @since  2004.06.03 */
/*--------------------------------------------------------------------*/
public class Font3D implements Serializable {

  private static final long serialVersionUID = 0x00010001L;

  /*------------------------------------------------------------------*/
  /*  constants                                                       */
  /*------------------------------------------------------------------*/
  /** direction flag: horizontal direction is positive x */
  public static final int H_POS_X = 0x00;
  /** direction flag: horizontal direction is positive y */
  public static final int H_POS_Y = 0x01;
  /** direction flag: horizontal direction is positive z */
  public static final int H_POS_Z = 0x02;
  /** direction flag: horizontal direction is negative x */
  public static final int H_NEG_X = 0x04;
  /** direction flag: horizontal direction is negative y */
  public static final int H_NEG_Y = 0x05;
  /** direction flag: horizontal direction is negative z */
  public static final int H_NEG_Z = 0x06;
  /** direction flag: vertical   direction is positive x */
  public static final int V_POS_X = 0x00;
  /** direction flag: vertical   direction is positive y */
  public static final int V_POS_Y = 0x10;
  /** direction flag: vertical   direction is positive z */
  public static final int V_POS_Z = 0x20;
  /** direction flag: vertical   direction is negative x */
  public static final int V_NEG_X = 0x40;
  /** direction flag: vertical   direction is negative y */
  public static final int V_NEG_Y = 0x50;
  /** direction flag: vertical   direction is negative z */
  public static final int V_NEG_Z = 0x60;

  /** the font vector table */
  private static final int font[] =
  { 0x82, 0x4a, 0x32, 0x39, 0x29, 0x2a, 0x3a, 0x71, 0x89, 0x83,
    0x7b, 0x60, 0x6c, 0x44, 0x48, 0x21, 0x89, 0x83, 0x2b, 0x30,
    0x29, 0x2b, 0x3c, 0x4c, 0x5b, 0x59, 0x68, 0x78, 0x89, 0x8b,
    0x7c, 0x92, 0x1a, 0x20, 0x8c, 0x61, 0x68, 0x88, 0x89, 0x69,
    0x44, 0x4b, 0x2b, 0x2c, 0x4c, 0x24, 0x3b, 0x4b, 0x5a, 0x59,
    0x68, 0x78, 0x89, 0x8a, 0x7b, 0x6b, 0x5a, 0x59, 0x48, 0x38,
    0x29, 0x2a, 0x3b, 0x4b, 0x5c, 0x61, 0x7a, 0x8a, 0x13, 0x39,
    0x79, 0x9b, 0x11, 0x3b, 0x7b, 0x99, 0x50, 0x5c, 0x74, 0x38,
    0x32, 0x7a, 0x70, 0x3c, 0x50, 0x5c, 0x72, 0x3a, 0x23, 0x2a,
    0x3a, 0x3b, 0x2b, 0x09, 0x50, 0x5c, 0x23, 0x2a, 0x3a, 0x3b,
    0x2b, 0x20, 0x8c, 0x30, 0x7c, 0x3c, 0x2b, 0x29, 0x38, 0x78,
    0x89, 0x8b, 0x7c, 0x60, 0x8a, 0x2a, 0x20, 0x2c, 0x70, 0x89,
    0x8b, 0x7c, 0x6c, 0x28, 0x2c, 0x70, 0x89, 0x8b, 0x7c, 0x6c,
    0x5b, 0x5a, 0x5b, 0x4c, 0x3c, 0x2b, 0x29, 0x38, 0x23, 0x8b,
    0x48, 0x4c, 0x84, 0x88, 0x58, 0x5b, 0x4c, 0x3c, 0x2b, 0x29,
    0x38, 0x50, 0x5b, 0x4c, 0x3c, 0x2b, 0x29, 0x38, 0x58, 0x7a,
    0x8c, 0x80, 0x8c, 0x28, 0x83, 0x7c, 0x6c, 0x5b, 0x59, 0x48,
    0x38, 0x29, 0x2b, 0x3c, 0x4c, 0x5b, 0x59, 0x68, 0x78, 0x89,
    0x8b, 0x20, 0x3a, 0x5c, 0x7c, 0x8b, 0x89, 0x78, 0x68, 0x59,
    0x5c, 0x33, 0x3a, 0x2a, 0x2b, 0x3b, 0x53, 0x6b, 0x6a, 0x5a,
    0x5b, 0x53, 0x5a, 0x6a, 0x6b, 0x5b, 0x23, 0x2a, 0x3a, 0x3b,
    0x2b, 0x09, 0x24, 0x59, 0x8c, 0x40, 0x4c, 0x60, 0x6c, 0x21,
    0x5c, 0x89, 0x70, 0x89, 0x8b, 0x7c, 0x6c, 0x4a, 0x32, 0x39,
    0x29, 0x2a, 0x3a, 0x43, 0x6b, 0x69, 0x49, 0x4b, 0x5c, 0x7c,
    0x8b, 0x89, 0x78, 0x38, 0x29, 0x2b, 0x3c, 0x20, 0x68, 0x8a,
    0x6c, 0x2c, 0x50, 0x5c, 0x20, 0x88, 0x8b, 0x7c, 0x6c, 0x5b,
    0x58, 0x20, 0x2b, 0x3c, 0x4c, 0x5b, 0x74, 0x8b, 0x89, 0x78,
    0x38, 0x29, 0x2b, 0x3c, 0x20, 0x88, 0x8b, 0x7c, 0x3c, 0x2b,
    0x28, 0x84, 0x88, 0x28, 0x2c, 0x50, 0x5a, 0x84, 0x88, 0x28,
    0x50, 0x5a, 0x74, 0x8b, 0x89, 0x78, 0x38, 0x29, 0x2b, 0x3c,
    0x5c, 0x5a, 0x80, 0x28, 0x24, 0x8c, 0x50, 0x5c, 0x20, 0x2c,
    0x22, 0x8a, 0x80, 0x8c, 0x84, 0x3c, 0x2b, 0x29, 0x38, 0x20,
    0x88, 0x84, 0x48, 0x51, 0x2c, 0x80, 0x28, 0x2c, 0x20, 0x88,
    0x5a, 0x8c, 0x2c, 0x20, 0x88, 0x2c, 0x8c, 0x83, 0x7c, 0x3c,
    0x2b, 0x29, 0x38, 0x78, 0x89, 0x8b, 0x20, 0x88, 0x8b, 0x7c,
    0x5c, 0x4b, 0x48, 0x83, 0x7c, 0x3c, 0x2b, 0x29, 0x38, 0x78,
    0x89, 0x8b, 0x42, 0x2c, 0x20, 0x88, 0x8b, 0x7c, 0x5c, 0x4b,
    0x48, 0x42, 0x2c, 0x30, 0x29, 0x2b, 0x3c, 0x4c, 0x5b, 0x59,
    0x68, 0x78, 0x89, 0x8b, 0x7c, 0x80, 0x8c, 0x82, 0x2a, 0x80,
    0x38, 0x29, 0x2b, 0x3c, 0x8c, 0x80, 0x2a, 0x8c, 0x80, 0x28,
    0x5a, 0x2c, 0x8c, 0x80, 0x2c, 0x20, 0x8c, 0x80, 0x78, 0x5a,
    0x2a, 0x52, 0x7c, 0x8c, 0x80, 0x8c, 0x28, 0x2c, 0x93, 0x99,
    0x19, 0x1b, 0x80, 0x2c, 0x91, 0x9b, 0x1b, 0x19, 0x60, 0x8a,
    0x6c, 0x10, 0x1c, 0x82, 0x7a, 0x6b, 0x61, 0x6b, 0x5c, 0x2c,
    0x29, 0x38, 0x49, 0x4c, 0x80, 0x28, 0x2b, 0x3c, 0x5c, 0x6b,
    0x68, 0x64, 0x69, 0x58, 0x38, 0x29, 0x2c, 0x64, 0x69, 0x58,
    0x38, 0x29, 0x2c, 0x8c, 0x40, 0x4c, 0x5c, 0x6b, 0x69, 0x58,
    0x38, 0x29, 0x2c, 0x84, 0x8b, 0x7a, 0x2a, 0x50, 0x5c, 0x00,
    0x0b, 0x1c, 0x6c, 0x69, 0x58, 0x38, 0x29, 0x2c, 0x80, 0x28,
    0x50, 0x69, 0x6b, 0x5c, 0x2c, 0x82, 0x7a, 0x60, 0x6a, 0x2a,
    0x20, 0x2c, 0x83, 0x7b, 0x61, 0x6b, 0x1b, 0x0a, 0x08, 0x80,
    0x28, 0x24, 0x48, 0x6c, 0x80, 0x8a, 0x2a, 0x20, 0x2c, 0x20,
    0x68, 0x69, 0x5a, 0x2a, 0x52, 0x6b, 0x5c, 0x2c, 0x20, 0x68,
    0x6b, 0x5c, 0x2c, 0x63, 0x69, 0x58, 0x38, 0x29, 0x2b, 0x3c,
    0x5c, 0x6b, 0x00, 0x68, 0x6b, 0x5c, 0x3c, 0x2b, 0x28, 0x04,
    0x6c, 0x69, 0x58, 0x38, 0x29, 0x2c, 0x60, 0x28, 0x40, 0x6a,
    0x6b, 0x5c, 0x20, 0x2b, 0x3c, 0x4b, 0x49, 0x58, 0x69, 0x6b,
    0x6c, 0x82, 0x3a, 0x2b, 0x2c, 0x60, 0x6c, 0x60, 0x38, 0x29,
    0x2c, 0x6c, 0x60, 0x2a, 0x6c, 0x60, 0x38, 0x29, 0x3a, 0x4a,
    0x3a, 0x2b, 0x3c, 0x6c, 0x60, 0x2c, 0x20, 0x6c, 0x60, 0x38,
    0x29, 0x2c, 0x00, 0x0b, 0x1c, 0x6c, 0x60, 0x6c, 0x28, 0x2c,
    0x94, 0x9b, 0x8a, 0x6a, 0x59, 0x4a, 0x2a, 0x1b, 0x1c, 0x82,
    0x2a, 0x90, 0x99, 0x8a, 0x6a, 0x5b, 0x4a, 0x2a, 0x19, 0x18,
    0x70, 0x89, 0x7b, 0x8c, 0x20, 0x2c, 0x6a, 0x28, 0x11, 0x0a,
    0x1b, 0x2a, 0x34, 0x2b, 0x29, 0x38, 0x78, 0x89, 0x8b, 0x7c,
    0x60, 0x38, 0x29, 0x2c, 0x6c, 0x71, 0x89, 0x83, 0x7b, 0x71,
    0x8b, 0x40, 0x4c, 0x5c, 0x6b, 0x69, 0x58, 0x38, 0x29, 0x2c,
    0x70, 0x9a, 0x7c, 0x61, 0x6b, 0x5c, 0x2c, 0x29, 0x38, 0x49,
    0x4c, 0x73, 0x8b, 0x81, 0x79, 0x61, 0x6b, 0x5c, 0x2c, 0x29,
    0x38, 0x49, 0x4c, 0x73, 0x89, 0x61, 0x6b, 0x5c, 0x2c, 0x29,
    0x38, 0x49, 0x4c, 0x72, 0x89, 0x9a, 0x8b, 0x7a, 0x61, 0x6b,
    0x5c, 0x2c, 0x29, 0x38, 0x49, 0x4c, 0x11, 0x0a, 0x1b, 0x2a,
    0x24, 0x29, 0x38, 0x58, 0x69, 0x6c, 0x70, 0x9a, 0x7c, 0x40,
    0x4c, 0x5c, 0x6b, 0x69, 0x58, 0x38, 0x29, 0x2c, 0x73, 0x8b,
    0x81, 0x79, 0x40, 0x4c, 0x5c, 0x6b, 0x69, 0x58, 0x38, 0x29,
    0x2c, 0x73, 0x89, 0x40, 0x4c, 0x5c, 0x6b, 0x69, 0x58, 0x38,
    0x29, 0x2c, 0x73, 0x8b, 0x81, 0x79, 0x60, 0x6a, 0x2a, 0x20,
    0x2c, 0x70, 0x9a, 0x7c, 0x60, 0x6a, 0x2a, 0x20, 0x2c, 0x73,
    0x89, 0x60, 0x6a, 0x2a, 0x20, 0x2c, 0x20, 0x68, 0x8a, 0x6c,
    0x2c, 0x83, 0x9b, 0x91, 0x89, 0x50, 0x5c, 0x20, 0x58, 0x8b,
    0x9a, 0x89, 0x5c, 0x2c, 0x40, 0x4c, 0x81, 0x9b, 0x74, 0x78,
    0x28, 0x2c, 0x50, 0x5a, 0x60, 0x69, 0x5a, 0x2a, 0x29, 0x38,
    0x49, 0x4c, 0x5c, 0x6b, 0x5a, 0x3a, 0x2b, 0x2c, 0x20, 0x68,
    0x89, 0x6a, 0x84, 0x8a, 0x2a, 0x2c, 0x50, 0x5c, 0x70, 0x9a,
    0x7c, 0x63, 0x69, 0x58, 0x38, 0x29, 0x2b, 0x3c, 0x5c, 0x6b,
    0x71, 0x89, 0x83, 0x7b, 0x63, 0x69, 0x58, 0x38, 0x29, 0x2b,
    0x3c, 0x5c, 0x6b, 0x81, 0x7b, 0x63, 0x69, 0x58, 0x38, 0x29,
    0x2b, 0x3c, 0x5c, 0x6b, 0x70, 0x9a, 0x7c, 0x64, 0x2c, 0x29,
    0x38, 0x68, 0x81, 0x7b, 0x64, 0x2c, 0x29, 0x38, 0x68, 0x71,
    0x89, 0x83, 0x7b, 0x64, 0x1c, 0x0b, 0x08, 0x24, 0x29, 0x38,
    0x68, 0x81, 0x99, 0x93, 0x8b, 0x73, 0x6c, 0x3c, 0x2b, 0x29,
    0x38, 0x68, 0x79, 0x7b, 0x81, 0x99, 0x93, 0x8b, 0x84, 0x3c,
    0x2b, 0x29, 0x38, 0x88, 0x30, 0x29, 0x2b, 0x3c, 0x5c, 0x6b,
    0x6b, 0x69, 0x58, 0x38, 0x20, 0x6c, 0x50, 0x5a, 0x83, 0x8a,
    0x79, 0x29, 0x28, 0x2b, 0x3c, 0x20, 0x8c, 0x74, 0x3c, 0x2b,
    0x29, 0x38, 0x78, 0x89, 0x8b, 0x7c, 0x20, 0x98, 0x9b, 0x8c,
    0x7c, 0x6b, 0x68, 0x42, 0x4c, 0x53, 0x2b, 0x2c, 0x10, 0x09,
    0x1a, 0x7a, 0x8b, 0x7c, 0x41, 0x4b, 0x71, 0x8b, 0x61, 0x6b,
    0x5c, 0x2c, 0x29, 0x38, 0x49, 0x4c, 0x71, 0x8b, 0x60, 0x6a,
    0x2a, 0x20, 0x2c, 0x71, 0x8b, 0x63, 0x69, 0x58, 0x38, 0x29,
    0x2b, 0x3c, 0x5c, 0x6b, 0x71, 0x8b, 0x64, 0x2c, 0x29, 0x38,
    0x68, 0x70, 0x89, 0x7b, 0x8c, 0x20, 0x68, 0x6b, 0x5c, 0x2c,
    0x80, 0x99, 0x8b, 0x9c, 0x20, 0x78, 0x2c, 0x7c, 0x81, 0x8b,
    0x7c, 0x4c, 0x49, 0x58, 0x69, 0x6c, 0x20, 0x2c, 0x20, 0x2c,
    0x43, 0x5c, 0x7c, 0x8b, 0x89, 0x78, 0x58, 0x49, 0x4b, 0x72,
    0x7b, 0x8b, 0x8a, 0x7a, 0x62, 0x48, 0x38, 0x29, 0x2b, 0x3c,
    0x20, 0x48, 0x4c, 0x40, 0x4c, 0x2c, 0x80, 0x99, 0x69, 0x20,
    0x9c, 0x42, 0x5b, 0x4c, 0x2a, 0x2c, 0x80, 0x99, 0x69, 0x20,
    0x9c, 0x23, 0x5b, 0x3a, 0x3c, 0x71, 0x7a, 0x8a, 0x89, 0x79,
    0x62, 0x2a, 0x72, 0x58, 0x3a, 0x34, 0x5a, 0x7c, 0x30, 0x5a,
    0x78, 0x72, 0x5c, 0x3a, 0x50, 0x5c, 0x2c, 0x28, 0x88, 0x8c,
    0x5c, 0x21, 0x89, 0x82, 0x2a, 0x23, 0x8b, 0x20, 0x2c, 0x5c,
    0x58, 0x28, 0x21, 0x59, 0x52, 0x2a, 0x23, 0x5b, 0x20, 0x2a,
    0x8a, 0x88, 0x28, 0x21, 0x89, 0x22, 0x2c, 0x8c, 0x8a, 0x2a,
    0x23, 0x8b, 0x50, 0x5c, 0x8c, 0x88, 0x58, 0x51, 0x89, 0x82,
    0x5a, 0x53, 0x8b, 0x64, 0x2a, 0x29, 0x38, 0x58, 0x69, 0x6a,
    0x2c, 0x00, 0x78, 0x89, 0x8b, 0x7c, 0x6c, 0x5b, 0x5a, 0x5b,
    0x4c, 0x3c, 0x2b, 0x28, 0x20, 0x88, 0x8c, 0x21, 0x69, 0x68,
    0x6c, 0x6b, 0x2b, 0x84, 0x88, 0x5a, 0x28, 0x2c, 0x51, 0x5a,
    0x4b, 0x3b, 0x2a, 0x29, 0x38, 0x48, 0x6a, 0x6c, 0x00, 0x19,
    0x69, 0x31, 0x2a, 0x2b, 0x3c, 0x6c, 0x40, 0x58, 0x69, 0x5a,
    0x1a, 0x42, 0x6c, 0x22, 0x8a, 0x73, 0x6c, 0x4c, 0x3b, 0x39,
    0x48, 0x68, 0x79, 0x7b, 0x51, 0x5b, 0x74, 0x8b, 0x89, 0x78,
    0x38, 0x29, 0x2b, 0x3c, 0x7c, 0x20, 0x29, 0x49, 0x58, 0x78,
    0x89, 0x8b, 0x7c, 0x5c, 0x4b, 0x2b, 0x2c, 0x52, 0x59, 0x48,
    0x38, 0x29, 0x2a, 0x3b, 0x4b, 0x78, 0x89, 0x8a, 0x7b, 0x30,
    0x58, 0x69, 0x5a, 0x3a, 0x2b, 0x3c, 0x5c, 0x6b, 0x5a, 0x3a,
    0x29, 0x38, 0x20, 0x6c, 0x63, 0x5c, 0x3c, 0x2b, 0x29, 0x38,
    0x58, 0x69, 0x6b, 0x50, 0x5a, 0x23, 0x29, 0x38, 0x78, 0x89,
    0x8b, 0x20, 0x78, 0x89, 0x8b, 0x7c, 0x2c, 0x70, 0x7c, 0x50,
    0x5c, 0x30, 0x3c, 0x20, 0x2c, 0x42, 0x8a, 0x60, 0x6c, 0x80,
    0x6c, 0x48, 0x20, 0x2c, 0x20, 0x2c, 0x44, 0x68, 0x8c, 0x02,
    0x8a, 0x9b, 0x8c, 0x10, 0x09, 0x1a, 0x9a, 0x50, 0x5b, 0x71,
    0x7a, 0x8a, 0x89, 0x79, 0x31, 0x29, 0x2a, 0x3a, 0x39, 0x50,
    0x69, 0x5b, 0x6c, 0x44, 0x3b, 0x49, 0x38, 0x51, 0x68, 0x78,
    0x89, 0x8a, 0x7b, 0x6b, 0x5a, 0x59, 0x51, 0x68, 0x78, 0x89,
    0x8a, 0x7b, 0x6b, 0x5a, 0x59, 0x61, 0x79, 0x7a, 0x6a, 0x69,
    0x30, 0x28, 0x29, 0x39, 0x38, 0x40, 0x2a, 0x9a, 0x9c, 0x8c,
    0x50, 0x88, 0x89, 0x7a, 0x5a, 0x70, 0x89, 0x7a, 0x58, 0x5a,
    0x80, 0x8a, 0x79, 0x6a, 0x59, 0x58, 0x80, 0x8c };

  /** the character offset table */
  private static final int offs[] =
  {    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    7,   11,   19,   33,   45,   65,
      68,   72,   76,   84,   88,   94,   96,  101,  103,  113,
     118,  125,  138,  142,  151,  161,  164,  181,  191,  201,
     212,  215,  219,  222,  233,  247,  254,  266,  274,  281,
     287,  292,  302,  308,  314,  319,  325,  328,  333,  337,
     346,  353,  364,  373,  385,  389,  395,  398,  403,  407,
     414,  418,  422,  424,  428,  431,  433,  436,  444,  451,
     457,  464,  473,  479,  488,  495,  502,  509,  514,  519,
     528,  533,  542,  549,  556,  562,  571,  577,  582,  585,
     594,  598,  606,  610,  619,  621,  630,  634,  638,  650,
     659,  670,  681,  693,  703,  716,  726,  738,  751,  762,
     771,  779,  786,  797,  806,  814,  828,  838,  850,  863,
     874,  882,  889,  901,  914,  924,  936,  945,  956,  968,
     976,  986,  993, 1004, 1011, 1020, 1028, 1038, 1049, 1060,
    1063, 1066, 1076, 1085, 1092, 1098, 1104, 1104, 1104, 1104,
    1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104,
    1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104,
    1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104,
    1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104, 1104,
    1117, 1128, 1135, 1142, 1153, 1161, 1174, 1177, 1183, 1188,
    1198, 1206, 1213, 1224, 1235, 1247, 1259, 1272, 1283, 1291,
    1297, 1303, 1309, 1314, 1319, 1323, 1327, 1339, 1347, 1356,
    1370, 1375, 1380, 1385, 1390, 1396, 1398 };

  /*------------------------------------------------------------------*/
  /*  instance variables                                              */
  /*------------------------------------------------------------------*/
  /** the 3d to 2d projection */
  private Proj3D   proj;
  /** the size of the font */
  private double   size;
  /** clipping: minimum and maximum in x-direction */
  private double   xmin, xmax;
  /** clipping: minimum and maximum in y-direction */
  private double   ymin, ymax;
  /** clipping: minimum and maximum in z-direction */
  private double   zmin, zmax;
  /** the buffer for the graphics */
  private transient Graphics g;
  /** the buffer for the projection */
  private double[] p;
  /** the buffer for the current point */
  private int      x, y;

  /*------------------------------------------------------------------*/
  /** Create a 3D font manager.
   *  @param  proj the 3D to 2D projection to use
   *  @since  2004.06.03 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public Font3D (Proj3D proj)
  {                             /* --- create a 3d font manager */
    this.proj = proj;           /* not the projection object */
    this.size = 0.1;            /* and initialize the variables */
    this.xmin = this.ymin = this.zmin = -Double.MAX_VALUE;
    this.xmax = this.ymax = this.zmax =  Double.MAX_VALUE;
    this.p    = new double[3];
  }  /* Font3D() */

  /*------------------------------------------------------------------*/
  /** Set the 3D to 2D projection to use.
   *  @param  proj the 3d to 2d projection to use
   *  @since  2004.06.03 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setProj3D (Proj3D proj)
  { this.proj = proj; }

  /*------------------------------------------------------------------*/
  /** Set the font size.
   *  @param  size the font size
   *  @since  2004.06.03 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setSize (double size)
  { this.size = 0.1 *size; }

  /*------------------------------------------------------------------*/
  /** Set the clipping box.
   *  @param  xmin the minimum in x-direction
   *  @param  xmax the maximum in x-direction
   *  @param  ymin the minimum in y-direction
   *  @param  ymax the maximum in y-direction
   *  @param  zmin the minimum in z-direction
   *  @param  zmax the maximum in z-direction
   *  @since  2004.06.06 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void setClip (double xmin, double xmax,
                       double ymin, double ymax,
                       double zmin, double zmax)
  {                             /* --- set the clipping block */
    this.xmin = xmin; this.xmax = xmax;
    this.ymin = ymin; this.ymax = ymax;
    this.zmin = zmin; this.zmax = zmax;
  }  /* setClip() */

  /*------------------------------------------------------------------*/
  /** Intersect/shrink the current clipping box.
   *  @param  xmin the minimum in x-direction
   *  @param  xmax the maximum in x-direction
   *  @param  ymin the minimum in y-direction
   *  @param  ymax the maximum in y-direction
   *  @param  zmin the minimum in z-direction
   *  @param  zmax the maximum in z-direction
   *  @since  2004.06.06 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void addClip (double xmin, double xmax,
                       double ymin, double ymax,
                       double zmin, double zmax)
  {                             /* --- shrink the clipping block */
    if (xmin > this.xmin) this.xmin = xmin;
    if (ymin > this.ymin) this.ymin = ymin;
    if (zmin > this.zmin) this.zmin = zmin;
    if (xmax < this.xmax) this.xmax = xmax;
    if (ymax < this.ymax) this.ymax = ymax;
    if (zmax < this.zmax) this.zmax = zmax;
  }  /* addClip() */

  /*------------------------------------------------------------------*/
  /** Move to or draw a line to a point.
   *  @param  x    the x-coordinate of the point
   *  @param  y    the y-coordinate of the point
   *  @param  z    the z-coordinate of the point
   *  @param  line whether to draw a line
   *  @since  2004.06.04 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  private void draw (double x, double y, double z, boolean line)
  {                             /* --- move to/line to a point */
    int nx, ny;                 /* new coordinates */

    if (this.proj.project(x, y, z, this.p) != 0) {
      this.x = Integer.MIN_VALUE; return; }
    nx =  (int)(this.p[0]+0.5); /* compute the coordinates */
    ny = -(int)(this.p[1]+0.5); /* of the new point */
    if ((x < this.xmin) || (x > this.xmax)
    ||  (y < this.ymin) || (y > this.ymax)
    ||  (z < this.zmin) || (z > this.zmax))
      nx = Integer.MIN_VALUE;   /* check whether point is valid */
    if (line                    /* compute projection and draw line */
    && (this.x > Integer.MIN_VALUE)
    && (nx     > Integer.MIN_VALUE))
      this.g.drawLine(this.x, this.y, nx, ny);
    this.x = nx; this.y = ny;   /* note the new point */
  }  /* draw() */

  /*------------------------------------------------------------------*/
  /** Draw a string at a given reference point.
   *  @param  g   the graphics to use for the drawing
   *  @param  x   the x-coordinate of the reference point
   *  @param  y   the y-coordinate of the reference point
   *  @param  z   the z-coordinate of the reference point
   *  @param  dir the drawing direction (given by flags)
   *  @param  s   the string to draw
   *  @since  2004.06.04 (Christian Borgelt) */
  /*------------------------------------------------------------------*/

  public void draw (Graphics g, double x, double y, double z, int dir,
                    String s)
  {                             /* --- draw a String */
    int    i, k;                /* loop variable */
    int    c, f;               /* current character/point */
    double dx, dy, dc;          /* offsets of current point */

    if ((s == null) || (this.size <= 0))
      return;                   /* check for a string to draw */
    this.g = g;                 /* note the graphics to draw with */
    dc = 6 *this.size;          /* compute the character step */
    if ((dir & 0x04) != 0) dc = -dc;
    for (i = 0; i < s.length(); i++) {
      c = s.charAt(i);          /* traverse the string */
      for (k = offs[c]; k < offs[c+1]; k++) {
        f  = font[k];           /* traverse the character's points */
        dy = this.size *((f >> 4) -2);
        dx = this.size * (f & 7);  /* get horiz. and vertical. offset */
        if ((dir & 0x04) != 0) dx = -dx;
        if ((dir & 0x40) != 0) dy = -dy;
        switch (dir & 0x33) {   /* evaluate the drawing direction */
          case 0x01: this.draw(x+dy, y+dx, z,    (f & 8) != 0); break;
          case 0x02: this.draw(x+dy, y,    z+dx, (f & 8) != 0); break;
          case 0x10: this.draw(x+dx, y+dy, z,    (f & 8) != 0); break;
          case 0x12: this.draw(x,    y+dy, z+dx, (f & 8) != 0); break;
          case 0x20: this.draw(x+dx, y,    z+dy, (f & 8) != 0); break;
          case 0x21: this.draw(x,    y+dx, z+dy, (f & 8) != 0); break;
        }                       /* compute coordinates of point */
      }                         /* in 3d space */
      switch (dir & 0x03) {     /* evaluate the direction of writing */
        case 0x00: x += dc; break;
        case 0x01: y += dc; break;
        case 0x02: z += dc; break;
      }                         /* advance the reference point */
    }                           /* of the character to draw */
  }  /* draw() */

}  /* Font3D */
